use diesel::QueryDsl;
#[cfg(not(target_env = "msvc"))]
use tikv_jemallocator::Jemalloc;

#[cfg(not(target_env = "msvc"))]
#[global_allocator]
static GLOBAL: Jemalloc = Jemalloc;

use actix_web::{get, http::StatusCode, middleware, web, App, HttpResponse, HttpServer};
use actix_web_static_files::ResourceFiles;
use askama_actix::{Template, TemplateToResponse};
use diesel::prelude::*;

use backend::*;

include!(concat!(env!("OUT_DIR"), "/generated.rs"));

mod filters {
    pub fn cmark<T: std::fmt::Display>(s: T) -> ::askama::Result<String> {
        let s = s.to_string();

        let options = pulldown_cmark::Options::empty();
        let parser = pulldown_cmark::Parser::new_ext(&s, options);

        let mut html_output = String::with_capacity(s.len() * 3 / 2);
        pulldown_cmark::html::push_html(&mut html_output, parser);

        Ok(html_output)
    }
}

#[derive(Template)]
#[template(path = "status_code.html")]
struct StatusCodeTemplate {
    status_code: StatusCode,
    message: Option<String>,
}

#[derive(Template)]
#[template(path = "posts/{slug}.html")]
struct PostBySlugTemplate {
    post: db::models::Post,
}

async fn not_found() -> HttpResponse {
    StatusCodeTemplate {
        status_code: StatusCode::NOT_FOUND,
        message: None,
    }
    .to_response()
}

#[get("/posts/{slug}")]
async fn post_by_slug(db_pool: web::Data<db::DbPool>, path: web::Path<String>) -> HttpResponse {
    let slug = path.into_inner();

    let conn = &mut match db_pool.get() {
        Ok(x) => x,
        Err(e) => return HttpResponse::InternalServerError().body(format!("{:?}", e)),
    };

    let post = match db::schema::posts::table
        .filter(db::schema::posts::slug.eq(&slug))
        .filter(db::schema::posts::active)
        .first::<db::models::Post>(conn)
        .optional()
    {
        Ok(x) => x,
        Err(e) => {
            dbg!(&e);
            return HttpResponse::InternalServerError().body(format!("{:?}", e));
        }
    };

    match post {
        Some(x) => PostBySlugTemplate { post: x }.to_response(),
        None => {
            let mut resp = StatusCodeTemplate {
                status_code: StatusCode::NOT_FOUND,
                message: None,
            }
            .to_response();
            *resp.status_mut() = StatusCode::NOT_FOUND;

            resp
        }
    }
}

#[actix_web::main]
async fn main() -> std::io::Result<()> {
    env_logger::init();
    log::info!("Listening to requests at http://0.0.0.0:80");

    HttpServer::new(move || {
        let generated = generate();

        App::new()
            .wrap(middleware::Compress::default())
            .wrap(middleware::Logger::default())
            .app_data(web::Data::new(db::pool().unwrap()))
            .service(post_by_slug)
            .service(ResourceFiles::new("/static", generated))
            .default_service(web::route().to(not_found))
    })
    .bind("0.0.0.0:80")
    .unwrap()
    .run()
    .await
}
