#[cfg(not(target_env = "msvc"))]
use tikv_jemallocator::Jemalloc;

#[cfg(not(target_env = "msvc"))]
#[global_allocator]
static GLOBAL: Jemalloc = Jemalloc;

use anyhow::{bail, Result};
use chrono::prelude::*;
use clap::{Parser, Subcommand};
use diesel::prelude::*;
use scan_dir::ScanDir;
use serde::{Deserialize, Serialize};
use std::fs;
use std::path::PathBuf;

use backend::*;

#[derive(Debug, Parser)]
#[clap(author, version, about, long_about = None)]
struct Cli {
    #[clap(subcommand)]
    commands: Commands,
}

#[derive(Debug, Subcommand)]
enum Commands {
    #[clap(about = "Imports new posts")]
    Import,
}

#[derive(Deserialize, Serialize, Debug)]
struct PostFrontmatter {
    id: Option<i32>,
    name: String,
    slug: String,
    description: String,
    published_at: NaiveDate,
    edited_at: Option<NaiveDate>,
    active: bool,
}

#[derive(Debug)]
struct Post {
    path: PathBuf,
    frontmatter: PostFrontmatter,
    content: String,
}

fn main() -> Result<()> {
    match Cli::parse().commands {
        Commands::Import => {
            let conn = &mut db::establish_connection()?;

            for post in ScanDir::dirs().read("/blog", |iter| {
                iter.map(|(entry, _)| {
                    let path = entry.path().join("post.md");
                    let src = fs::read_to_string(&path)?;
                    let frontmatter = match fronma::parser::parse::<PostFrontmatter>(&src) {
                        Ok(x) => x,
                        Err(x) => bail!("Error parsing frontmatter: {:?}", x),
                    };

                    Ok(Post {
                        path,
                        frontmatter: frontmatter.headers,
                        content: frontmatter.body.to_string(),
                    })
                })
                .collect::<Result<Vec<_>>>()
            })?? {
                let content = post.content.trim();

                if let Some(id) = post.frontmatter.id {
                    diesel::update(db::schema::posts::table)
                        .filter(db::schema::posts::id.eq(id))
                        .set(db::models::UpdatePost {
                            name: Some(&post.frontmatter.name),
                            slug: Some(&post.frontmatter.slug),
                            description: Some(&post.frontmatter.description),
                            content: Some(content),
                            published_at: Some(post.frontmatter.published_at),
                            edited_at: Some(post.frontmatter.edited_at),
                            active: Some(post.frontmatter.active),
                        })
                        .execute(conn)?;
                } else {
                    let id = if let Some(id) = db::schema::posts::table
                        .select(db::schema::posts::id)
                        .filter(db::schema::posts::slug.eq(&post.frontmatter.slug))
                        .first::<i32>(conn)
                        .optional()?
                    {
                        diesel::update(db::schema::posts::table)
                            .filter(db::schema::posts::id.eq(id))
                            .set(db::models::UpdatePost {
                                name: Some(&post.frontmatter.name),
                                slug: None,
                                description: Some(&post.frontmatter.description),
                                content: Some(content),
                                published_at: Some(post.frontmatter.published_at),
                                edited_at: Some(post.frontmatter.edited_at),
                                active: Some(post.frontmatter.active),
                            })
                            .execute(conn)?;

                        id
                    } else {
                        diesel::insert_into(db::schema::posts::table)
                            .values(db::models::NewPost {
                                name: &post.frontmatter.name,
                                slug: &post.frontmatter.slug,
                                description: &post.frontmatter.description,
                                content: content,
                                published_at: post.frontmatter.published_at,
                                edited_at: post.frontmatter.edited_at,
                                active: post.frontmatter.active,
                            })
                            .returning(db::schema::posts::id)
                            .get_result::<i32>(conn)?
                    };

                    fs::write(
                        post.path,
                        format!(
                            "---\n{}---\n{}",
                            serde_yaml::to_string(&PostFrontmatter {
                                id: Some(id),
                                ..post.frontmatter
                            })?,
                            post.content
                        ),
                    )?;
                }
            }

            Ok(())
        }
    }
}
